from .freedman_diaconis import threshold_freedman_diaconis
from .scott import threshold_scott
from .sturges import threshold_sturges

__all__ = [
    "threshold_freedman_diaconis",
    "threshold_scott",
    "threshold_sturges",
]
